#!/usr/bin/env python

# DSR:  begin block 2002.01.19
# The core mx modules were restructured recently (2001?) and grouped under the
# title "eGenix mx Base Package".  Rather than standing alone, modules such as
# DateTime now belong to a package named "mx".
# On systems with the restructured mx packages, "import DateTime" will not
# work; "from mx import DateTime" should be used instead.

# The following code first tries the new import style, then the old style if
# the new style fails.
try:
    from mx import DateTime # new style
except ImportError:
    import DateTime # old style
# DSR:  end block 2002.01.19

# DSR:  begin block 2002.01.19
# Importing the string module is necessary for a Python 1.5.2-compatible fix
# for bug #438130 ("cursor.callproc not adding param code").
import string
# DSR:  end block 2002.01.19

import _kinterbasdb
import exceptions
import types
import UserList

#error = _kinterbasdb.error
#dbi = _kinterbasdb.dbi

apilevel = '2.0'
threadsafety = 1
paramstyle = 'qmark'

Warning = _kinterbasdb.Warning
Error = _kinterbasdb.Error
InterfaceError = _kinterbasdb.InterfaceError
DatabaseError = _kinterbasdb.DatabaseError
DataError = _kinterbasdb.DataError
OperationalError = _kinterbasdb.OperationalError
IntegrityError = _kinterbasdb.IntegrityError
InternalError = _kinterbasdb.InternalError
ProgrammingError = _kinterbasdb.ProgrammingError
NotSupportedError = _kinterbasdb.NotSupportedError

Date = DateTime.DateTime
Time = DateTime.TimeDelta
Timestamp = DateTime.DateTime
DateFromTicks = DateTime.DateFromTicks
TimeFromTicks = DateTime.TimeFromTicks
TimestampFromTicks = DateTime.localtime
Binary = buffer

class DBAPITypeObject:
    def __init__(self, *values):
        self.values  =  values
    def __cmp__(self, other):
        if other in self.values:
            return 0
        if other < self.values:
            return 1
        else:
            return -1

STRING = DBAPITypeObject(types.StringType)
BINARY = DBAPITypeObject(types.BufferType)
NUMBER = DBAPITypeObject(types.IntType, types.LongType, types.FloatType)
DATETIME = DBAPITypeObject(DateTime.DateTimeType, DateTime.DateTimeDelta)
ROWID = DBAPITypeObject()

def connect(*arg, **kw):
    return apply(ConnectionObject, arg, kw)

def execute_immediate(sql):
    """Fast execution of simple statements"""
    _kinterbasdb.execute_immediate(None, sql)



######################################################
#        Below is some constants
######################################################

# transaction parameters
isc_tpb_consistency               =  '\001'
isc_tpb_concurrency               =  '\002'
isc_tpb_shared                    =  '\003'
isc_tpb_protected                 =  '\004'
isc_tpb_exclusive                 =  '\005'
isc_tpb_wait                      =  '\006'
isc_tpb_nowait                    =  '\007'
isc_tpb_read                      =  '\010'
isc_tpb_write                     =  '\011'
isc_tpb_lock_read                 =  '\012'
isc_tpb_lock_write                =  '\013'
isc_tpb_verb_time                 =  '\014'
isc_tpb_commit_time               =  '\015'
isc_tpb_ignore_limbo              =  '\016'
isc_tpb_read_committed            =  '\017'
isc_tpb_autocommit                =  '\020'
isc_tpb_rec_version               =  '\021'
isc_tpb_no_rec_version            =  '\022'
isc_tpb_restart_requests          =  '\023'
isc_tpb_no_auto_undo              =  '\024'

# default transaction parameters
default_tpb  =  ( isc_tpb_concurrency +
                isc_tpb_shared +
                isc_tpb_wait +
                isc_tpb_write +
                isc_tpb_read_committed +
                isc_tpb_rec_version )

################################
## Database information items ##
################################

isc_info_db_id                    =  4
isc_info_reads                    =  5
isc_info_writes                   =  6
isc_info_fetches                  =  7
isc_info_marks                    =  8
isc_info_implementation           =  11
isc_info_version                  =  12
isc_info_base_level               =  13
isc_info_page_size                =  14
isc_info_num_buffers              =  15
isc_info_limbo                    =  16
isc_info_current_memory           =  17
isc_info_max_memory               =  18
isc_info_window_turns             =  19
isc_info_license                  =  20
isc_info_allocation               =  21
isc_info_attachment_id            =  22
isc_info_read_seq_count           =  23
isc_info_read_idx_count           =  24
isc_info_insert_count             =  25
isc_info_update_count             =  26
isc_info_delete_count             =  27
isc_info_backout_count            =  28
isc_info_purge_count              =  29
isc_info_expunge_count            =  30
isc_info_sweep_interval           =  31
isc_info_ods_version              =  32
isc_info_ods_minor_version        =  33
isc_info_no_reserve               =  34
isc_info_logfile                  =  35
isc_info_cur_logfile_name         =  36
isc_info_cur_log_part_offset      =  37
isc_info_num_wal_buffers          =  38
isc_info_wal_buffer_size          =  39
isc_info_wal_ckpt_length          =  40
isc_info_wal_cur_ckpt_interval    =  41
isc_info_wal_prv_ckpt_fname       =  42
isc_info_wal_prv_ckpt_poffset     =  43
isc_info_wal_recv_ckpt_fname      =  44
isc_info_wal_recv_ckpt_poffset    =  45
isc_info_wal_grpc_wait_usecs      =  47
isc_info_wal_num_io               =  48
isc_info_wal_avg_io_size          =  49
isc_info_wal_num_commits          =  50
isc_info_wal_avg_grpc_size        =  51
isc_info_forced_writes            =  52
isc_info_user_names               =  53
isc_info_page_errors              =  54
isc_info_record_errors            =  55
isc_info_bpage_errors             =  56
isc_info_dpage_errors             =  57
isc_info_ipage_errors             =  58
isc_info_ppage_errors             =  59
isc_info_tpage_errors             =  60
isc_info_set_page_buffers         =  61
isc_info_db_SQL_dialect           =  62
isc_info_db_read_only             =  63
isc_info_db_size_in_pages         =  64


class ConnectionObject:
    def __init__(self, *arg, **kw):
        self._conn = apply(_kinterbasdb.attach_db, arg, kw)
        self.server_version = _kinterbasdb.server_version(self._conn)
        self.default_tpb = default_tpb

    def begin(self, param = None):
        """start transaction"""
        if param:
            _kinterbasdb.begin(self._conn, param)
        else:
            if self.server_version<4:
                # if connected to IB 3
                _kinterbasdb.begin(self._conn)
            else:
                _kinterbasdb.begin(self._conn, self.default_tpb)

    def commit(self):
        """commit transaction"""
        _kinterbasdb.commit(self._conn)

    def rollback(self):
        """commit transaction"""
        _kinterbasdb.rollback(self._conn)

    def execute_immediate(self, sql):
        """fast execution"""
        _kinterbasdb.execute_immediate(self._conn, sql)

    def cursor(self):
        """get cursor object"""
        return Cursor(self)

    def close(self):
        """closes connection to server"""
        if hasattr(self, "_conn"):
            del self._conn

    def __getattr__(self, attr):
        if attr == "dialect":
            return _kinterbasdb.get_dialect(self._conn)
        else:
            return self.__dict__[attr]

    def __setattr__(self, attr, value):
        if attr == "dialect":
            return _kinterbasdb.set_dialect(self._conn, value)
        else:
            self.__dict__[attr] = value

    def database_info(self, request, type):
        """Wrapper for isc_database_info.

Parameters
    request: use isc_info_* to say what info do you need
    type: tells what type of result is expected. 's'- string or 'i'-short integer."""
        return _kinterbasdb.database_info(self._conn, request, type)


class FieldSet(UserList.UserList):
    def __init__(self, description, data):
        if len(description) != len(data):
            raise InternalError, "cursor description does not match result"
        UserList.UserList.__init__(self, data)
        self.description = description

    def __getitem__(self, key):
        if type(key) is types.StringType:
            for i in range(len(self.description)):
                if key == self.description[i][0]:
                    return self.data[i]
        else:
            return self.data[key]
        raise exceptions.KeyError, key

    def __str__(self):
        res = ''
        for i in range(len(self.data)):
            if self.description[i][1] == DATETIME and \
               self.description[i][2] == 10:
                s = '%-*.*s'%(self.description[i][2], self.description[i][2], str(self.data[i]))
            else:
                s = '%-*s'%(self.description[i][2], str(self.data[i]))
            res = res+s+' '
        return res;

class DescriptionField(UserList.UserList):
    def __init__(self, description):
        UserList.UserList.__init__(self, description)
        self.description = description

    def __getattr__(self, key):
        if key == 'name':
            return self.description[0]
        elif key == 'type_code':
            return self.description[1]
        elif key == 'display_size':
            return self.description[2]
        elif key == 'internal_size':
            return self.description[3]
        elif key == 'precision':
            return self.description[4]
        elif key == 'scale':
            return self.description[5]
        elif key == 'null_ok':
            return self.description[6]
        raise exceptions.KeyError, key

class Description(UserList.UserList):
    def __init__(self, description):
        l = []
        for d in description:
            l.append(DescriptionField(d))
        UserList.UserList.__init__(self, l)

    def __getitem__(self, key):
        if type(key) is types.StringType:
            for d in self.data:
                if key == d[0]:
                    return d
        else:
            return self.data[key]
        raise exceptions.KeyError, key

    def __str__(self):
        res = ''
        for d in self.data:
            res = res+'%-*s '%(d[2], d[0])
        return res;




class Cursor:
    def __init__(self, _conn):
        self._conn = _conn;
        self.cur = _kinterbasdb.cursor(self._conn._conn)
        self.description = ()

    def __del__( self ):
        if hasattr( self, "_conn" ):
            del self._conn
        if hasattr( self, "cur" ):
            del self.cur
        if hasattr( self, "description" ):
            del self.description

    def execute(self, sql, params = None):
        res = (-1, ())
        if params == None:
            res = _kinterbasdb.execute(self.cur, sql)
        else:
            res = _kinterbasdb.execute(self.cur, sql, params)
        self.description = Description(res[1])
        return res[0]

    def executemany(self, sql, params = None):
        res = (-1, ())
        if params == None:
            res = _kinterbasdb.execute(self.cur, sql)
        else:
            for param in params:
                res = _kinterbasdb.execute(self.cur, sql, param)
        self.description = Description(res[1])
        return res[0]

    def fetch(self):
        res = _kinterbasdb.fetch(self.cur)
        if not res: return None
        return FieldSet(self.description, list(res))

    def fetchone(self):
        return self.fetch()

    def fetchall(self):
        res = None
        all = []
        while 1:
            res = _kinterbasdb.fetch( self.cur )
            if res is None:
                break
            all.append(res)
        return all

    def fetchmany(self, count):
        res = None
        all = []
        i = 0
        while i<count:
            res = _kinterbasdb.fetch(self.cur)
            if res == None:
                break
            all.append(res)
# DSR:  begin block 2002.01.19
# Bug #468304 reported and fixed by Stjepan Zlodi ("vuk"):
            i = i + 1
# DSR:  end block 2002.01.19
        return all

    def close(self):
        return _kinterbasdb.close(self.cur)

# DSR:  begin block 2002.01.19
# Bug #438130 fix:
    def callproc(self, procname, params=None):
        sql = 'EXECUTE PROCEDURE ' + procname
        if params is not None:
            sql = sql + ' ' + string.join(('?',) * len(params), ',')
        return self.execute(sql, params)
# DSR:  end block 2002.01.19

def test():
    try:
        import time
        help = '/usr/interbase/help/help.gdb'
        print '*'*8+'Test for kinterbasdb module'+'*'*8
        print("Connecting to "+ help)
        user = raw_input("user: ")
        password = raw_input("password: ")
        c = connect(help, user, password)
        print "Connected Ok", c._conn, '(server: interbase ', c.server_version,')'
        if c.server_version >= 6:
            print "Database SQL dialect: ", c.database_info(isc_info_db_SQL_dialect, 'i')
            print "License: ", c.database_info(isc_info_license, 's')
            c.dialect = 3
            print "Client SQL dialect set to", c.dialect
        a = c.cursor()
        print "cursor ok", a.cur
        c.begin()
        a.execute('select * from TOPICS where topic  =  ?', ('BEGIN', ))
        print 'description:', repr(a.description)
        print "execute ok", a.cur
        while 1:
            res = a.fetch()
            if not res: break
        print "fetch ok"
        a.close()
        print "close ok"
        c.commit()
        print "Ok"
    except Error, e:
        print Error, ":", e

def test_gen_win32_ib6():
    """ This implied that the Interbase ver. 6.0 
        was installed with examples to
        path  =  "c:\Program Files\Borland\InterBase\Examples"
    """
    IB6DIR = "C:\\Program Files\\Borland\\InterBase"
    PATH_TO_DATABASE = IB6DIR+"\\Examples\\Database"
    DATABASE = "employee.gdb"
    DATA = PATH_TO_DATABASE+"\\"+DATABASE
    try:
        import time
        print '*'*8+'Test for kinterbasdb module'+'*'*8
        print("Connecting to "+ DATA)
        user = raw_input("\nuser (SYSDBA, case insensitive): ")
        password = raw_input("\npassword (masterkey, case sensitive,\n"
                "if it was not changed since installation of \n"
                "Interbase): ")
        c = connect(dsn = DATA, user = user, password = password)
        print "Connected Ok", c._conn, '(server: interbase ', c.server_version,')'
        if c.server_version >= 6:
            print "Database SQL dialect: ", c.database_info(isc_info_db_SQL_dialect, 'i')
            print "License: ", c.database_info(isc_info_license, 's')
            c.dialect = 3
            print "Client SQL dialect set to", c.dialect
        a = c.cursor()
        print "cursor ok", a.cur
        c.begin()
        a.execute("""select FIRST_NAME, LAST_NAME, JOB_COUNTRY,
                     HIRE_DATE FROM employee 
                     WHERE JOB_COUNTRY  =  ?""", ('USA', ))
        print 'description:', repr(a.description)
        print "execute ok", a.cur
        while 1:
            res = a.fetch()
            if not res: break
            print res
        print "fetch ok"
        a.close()
        print "close ok"
        c.commit()
        print "Ok"
    except Error, e:
        print Error, ":", e

if __name__ == '__main__':
    import os
    if os.name == "nt":
        test_gen_win32_ib6()
    else:
        # Run old test
        test()
